from .base_engine import BaseEngine
from elevenlabs import voices, generate, stream
from elevenlabs.api import Voice, VoiceSettings
import pyaudio
import itertools
from typing import Iterator
import subprocess
import threading
import shutil
import subprocess
import elevenlabs
from typing import Union

class ElevenlabsVoice:
    def __init__(self, name, voice_id, category, description, labels):
        self.name = name
        self.voice_id = voice_id
        self.category = category
        self.description = description
        self.labels = labels

    def __repr__(self):
        label_string = ', '.join(self.labels.values())
        #return f"{self.name} ({self.category}, {label_string})"
        return f"{self.name} ({self.category}, id: {self.voice_id}, {label_string})"
        #return f"<Voice(name={self.name}, locale={self.locale}, gender={self.gender})>"


class ElevenlabsEngine(BaseEngine):

    def __init__(self, 
                 api_key: str = "", 
                 voice: str = "Nicole", 
                 id: str = "piTKgcLEGmPE4e6mEKli",
                 category: str = "premade",
                 clarity: float = 100.0,
                 stability: float = 30.0,
                 model: str = "eleven_multilingual_v1"):
        """
        Initializes an elevenlabs voice realtime text to speech engine object.

        Args:
            api_key (str): Elevenlabs API key. (TTS API key)
            voice (str, optional): Voice name. Defaults to "Nicole".
            id (str, optional): Voice ID. Defaults to "piTKgcLEGmPE4e6mEKli".
            category (str, optional): Voice category. Defaults to "premade".
            clarity (float, optional): Clarity. Adjusts voice similarity and resemblance. Defaults to "100.0".
            stability (float, optional): Stability. Controls the voice performance, with higher values producing a steadier tone and lower values giving a more emotive output. Defaults to "30.0".
            
        """

        self.voice_name = voice
        self.id = id
        self.category = category
        self.clarity = clarity
        self.stability = stability
        self.model = model
        self.pause_event = threading.Event()
        self.immediate_stop = threading.Event()

        self.set_api_key(api_key)
        
    def post_init(self):
        """ Information that this engine can handle generators directly """ 
        self.can_consume_generators = True

    def get_stream_info(self):
        """
        Returns the audio stream configuration information suitable for PyAudio.

        Returns:
            tuple: A tuple containing the audio format, number of channels, and the sample rate.
                  - Format (int): The format of the audio stream. pyaudio.paInt16 represents 16-bit integers.
                  - Channels (int): The number of audio channels. 1 represents mono audio.
                  - Sample Rate (int): The sample rate of the audio in Hz. 16000 represents 16kHz sample rate.
        """        
        return pyaudio.paInt16, 1, 16000
    
    def pause(self):
        """Pauses playback of the synthesized audio stream (won't work properly with elevenlabs)."""
        self.pause_event.set()

    def resume(self):
        """Resumes a previously paused playback of the synthesized audio stream (won't work properly with elevenlabs)."""
        self.pause_event.clear()

    def stop(self):
        """Stops the playback of the synthesized audio stream immediately."""
        self.mpv_process.terminate()
        return True

    def synthesize(self, 
                   generator: Iterator[str]):
        """
        Synthesizes text to audio stream.

        Args:
            text (str): Text to synthesize.
        """

        self.immediate_stop.clear()

        first_chunk = next(generator)
        combined_gen = itertools.chain([first_chunk], generator)

        voice_object = Voice(
            voice_id=self.id,
            name=self.voice_name,
            category=self.category,
            settings=VoiceSettings(stability=self.stability/100, similarity_boost=self.clarity/100)
        )

        self.audio_stream = generate(
            text=combined_gen,
            voice=voice_object,
            model=self.model,
            stream=True
        )

        self.stream(self.audio_stream)
        
    def set_api_key(self, api_key: str):
        """
        Sets the elevenlabs api key. 

        Args:
            api_key (str): Elevenlabs API key. (TTS API key)
        """
        self.api_key = api_key
        if api_key: 
            elevenlabs.set_api_key(api_key)
    
    def stream(self, audio_stream: Iterator[bytes]) -> bytes:
        """
        Stream the audio data using the 'mpv' player.

        This method takes the audio_stream iterator, which contains bytes of audio data,
        and plays them using the 'mpv' player. The function will continuously feed the
        'mpv' process with chunks of audio data until the audio is finished or an error
        occurs.

        If the 'mpv' player is not installed, a ValueError is raised.

        Args:
            audio_stream (Iterator[bytes]): An iterator that yields bytes of audio data.

        Returns:
            bytes: The concatenated bytes of the entire audio stream.

        Raises:
            ValueError: If the 'mpv' player is not found.
        """        
        if not self.is_installed("mpv"):
            message = (
                "mpv not found, necessary to stream audio. "
                "On mac you can install it with 'brew install mpv'. "
                "On linux and windows you can install it from https://mpv.io/"
            )
            raise ValueError(message)

        mpv_command = ["mpv", "--no-cache", "--no-terminal", "--", "fd://0"]
        self.mpv_process = subprocess.Popen(
            mpv_command,
            stdin=subprocess.PIPE,
            stdout=subprocess.DEVNULL,
            stderr=subprocess.DEVNULL,
        )

        audio = b""

        try:
            for chunk in audio_stream:
                if chunk is not None:
                    self.mpv_process.stdin.write(chunk)
                    self.mpv_process.stdin.flush()
                    audio += chunk
        except BrokenPipeError:
            # broken pipe error if mpv is terminated prematurely
            # we need to handle this because we terminate the process in stop method
            pass

        if self.mpv_process.stdin:
            self.mpv_process.stdin.close()
        self.mpv_process.wait()

        return audio
    
    def is_installed(self, lib_name: str) -> bool:
        """
        Check if the given library or software is installed and accessible.

        This method uses shutil.which to determine if the given library or software is
        installed and available in the system's PATH.

        Args:
            lib_name (str): Name of the library or software to check.

        Returns:
            bool: True if the library is installed, otherwise False.
        """        
        lib = shutil.which(lib_name)
        if lib is None:
            return False
        return True    
    
    def get_voices(self):
        """
        Retrieves the voices available from the Elevenlabs voice source.

        Calling this takes time, it sends a request to the Elevenlabs API to fetch the list of available voices.
        This method fetches the list of available voices using the elevenlabs `voices()` function and then
        constructs a list of `ElevenlabsVoice` objects to represent each voice's details.       

        Returns:
            list[ElevenlabsVoice]: A list containing ElevenlabsVoice objects representing each available voice. 
                                Each ElevenlabsVoice object encapsulates information such as the voice's name, 
                                ID, category, description, and associated labels.

        Note:
            This method relies on the `voices()` function to obtain the raw voice data. Ensure that the 
            `voices()` function is accessible and functional before calling this method.
        """        
        fetched_voices = voices()

        voice_objects = []
        for voice in fetched_voices:
            voice_object = ElevenlabsVoice(voice.name, voice.voice_id, voice.category, voice.description, voice.labels)
            voice_objects.append(voice_object)
        return voice_objects
    
    def set_voice(self, voice: Union[str, ElevenlabsVoice]):
        """
        Sets the voice to be used for speech synthesis.

        Args:
            voice (Union[str, ElevenlabsVoice]): The voice to be used for speech synthesis.
        """
        if isinstance(voice, ElevenlabsVoice):
            self.voice_name = voice.name
            self.id = voice.voice_id
            self.category = voice.category
        else:
            installed_voices = self.get_voices()
            for installed_voice in installed_voices:
                if voice.name in installed_voice.name:
                    self.voice_name = installed_voice.name
                    self.id = installed_voice.voice_id
                    self.category = installed_voice.category


"""
Stream management
"""

import pyaudio
import queue
import threading
import logging
import time


class AudioConfiguration:
    """
    Defines the configuration for an audio stream.
    """

    def __init__(self, format: int = pyaudio.paInt16, channels: int = 1, rate: int = 16000):
        """
        Args:
            format (int): Audio format, defaults to pyaudio.paInt16
            channels (int): Number of channels, defaults to 1 (mono)
            rate (int): Sample rate, defaults to 16000
        """
        self.format = format
        self.channels = channels
        self.rate = rate


class AudioStream:
    """
    Handles audio stream operations such as opening, starting, stopping, and closing.
    """

    def __init__(self, config: AudioConfiguration):
        """
        Args:
            config (AudioConfiguration): Object containing audio settings.
        """
        self.config = config
        self.stream = None
        self.pyaudio_instance = pyaudio.PyAudio()

    def open_stream(self):
        """Opens an audio stream."""
        self.stream = self.pyaudio_instance.open(format=self.config.format, channels=self.config.channels, rate=self.config.rate, output=True)

    def start_stream(self):
        """Starts the audio stream."""
        if self.stream and not self.stream.is_active():
            self.stream.start_stream()

    def stop_stream(self):
        """Stops the audio stream."""
        if self.stream and self.stream.is_active():
            self.stream.stop_stream()

    def close_stream(self):
        """Closes the audio stream."""
        if self.stream:
            self.stop_stream()
            self.stream.close()

    def is_stream_active(self) -> bool:
        """
        Checks if the audio stream is active.

        Returns:
            bool: True if the stream is active, False otherwise.
        """
        return self.stream and self.stream.is_active()


class AudioBufferManager:
    """
    Manages an audio buffer, allowing addition and retrieval of audio data.
    """

    def __init__(self, audio_buffer: queue.Queue):
        """
        Args:
            audio_buffer (queue.Queue): Queue to be used as the audio buffer.
        """
        self.audio_buffer = audio_buffer
        self.total_samples = 0

    def add_to_buffer(self, audio_data):
        """
        Adds audio data to the buffer.

        Args:
            audio_data: Audio data to be added.
        """
        self.audio_buffer.put(audio_data)
        self.total_samples += len(audio_data) // 2

    def clear_buffer(self):
        """Clears all audio data from the buffer."""
        while not self.audio_buffer.empty():
            try:
                self.audio_buffer.get_nowait()
            except queue.Empty:
                continue
        self.total_samples = 0

    def get_from_buffer(self, timeout: float = 0.05):
        """
        Retrieves audio data from the buffer.

        Args:
            timeout (float): Time (in seconds) to wait before raising a queue.Empty exception.

        Returns:
            The audio data chunk or None if the buffer is empty.
        """
        try:
            chunk = self.audio_buffer.get(timeout=timeout)
            self.total_samples -= len(chunk) // 2
            return chunk
        except queue.Empty:
            return None

    def get_buffered_seconds(self, rate: int) -> float:
        """
        Calculates the duration (in seconds) of the buffered audio data.

        Args:
            rate (int): Sample rate of the audio data.

        Returns:
            float: Duration of buffered audio in seconds.
        """
        return self.total_samples / rate


class StreamPlayer:
    """
    Manages audio playback operations such as start, stop, pause, and resume.
    """

    def __init__(self, audio_buffer: queue.Queue, config: AudioConfiguration):
        """
        Args:
            audio_buffer (queue.Queue): Queue to be used as the audio buffer.
            config (AudioConfiguration): Object containing audio settings.
        """
        self.buffer_manager = AudioBufferManager(audio_buffer)
        self.audio_stream = AudioStream(config)
        self.playback_active = False
        self.immediate_stop = threading.Event()
        self.pause_event = threading.Event()
        self.playback_thread = None

    def _play_chunk(self, chunk):
        """
        Plays a chunk of audio data.

        Args:
            chunk: Chunk of audio data to be played.
        """
        sub_chunk_size = 1024
        for i in range(0, len(chunk), sub_chunk_size):
            sub_chunk = chunk[i:i + sub_chunk_size]
            self.audio_stream.stream.write(sub_chunk)

            # Pause playback if the event is set
            while self.pause_event.is_set():
                time.sleep(0.01)

            if self.immediate_stop.is_set():
                break

    def _process_buffer(self):
        """Processes and plays audio data from the buffer until it's empty or playback is stopped."""
        while self.playback_active or not self.buffer_manager.audio_buffer.empty():
            chunk = self.buffer_manager.get_from_buffer()
            if chunk:
                self._play_chunk(chunk)

            if self.immediate_stop.is_set():
                logging.info("Immediate stop requested, aborting playback")
                break

    def get_buffered_seconds(self) -> float:
        """
        Calculates the duration (in seconds) of the buffered audio data.

        Returns:
            float: Duration of buffered audio in seconds.
        """
        total_samples = sum(len(chunk) // 2 for chunk in list(self.buffer_manager.audio_buffer.queue))
        return total_samples / self.audio_stream.config.rate

    def start(self):
        """Starts audio playback."""
        self.playback_active = True
        self.audio_stream.open_stream()
        self.audio_stream.start_stream()
        self.playback_thread = threading.Thread(target=self._process_buffer)
        self.playback_thread.start()

    def stop(self, immediate: bool = False):
        """
        Stops audio playback.

        Args:
            immediate (bool): If True, stops playback immediately without waiting for buffer to empty.
        """
        if not self.playback_thread:
            logging.warn("No playback thread found, cannot stop playback")
            return

        if immediate:
            self.immediate_stop.set()
            return

        self.playback_active = False

        if self.playback_thread and self.playback_thread.is_alive():
            self.playback_thread.join()

        time.sleep(0.1)

        self.audio_stream.close_stream()
        self.immediate_stop.clear()
        self.buffer_manager.clear_buffer()
        self.playback_thread = None

    def pause(self):
        """Pauses audio playback."""
        self.pause_event.set()

    def resume(self):
        """Resumes paused audio playback."""
        self.pause_event.clear()


# import pyaudio
# import queue
# import threading
# import logging
# import time

# class AudioConfiguration:
#     def __init__(self, format: int = pyaudio.paInt16, channels: int = 1, rate: int = 16000):
#         self.format = format
#         self.channels = channels
#         self.rate = rate


# class AudioStream:
#     def __init__(self, config: AudioConfiguration):
#         self.config = config
#         self.stream = None
#         self.pyaudio_instance = pyaudio.PyAudio()

#     def open_stream(self):
#         self.stream = self.pyaudio_instance.open(format=self.config.format, channels=self.config.channels, rate=self.config.rate, output=True)
    
#     def start_stream(self):
#         if self.stream and not self.stream.is_active():
#             self.stream.start_stream()

#     def stop_stream(self):
#         if self.stream and self.stream.is_active():
#             self.stream.stop_stream()

#     def close_stream(self):
#         if self.stream:
#             self.stop_stream()
#             self.stream.close()

#     def is_stream_active(self):
#         return self.stream and self.stream.is_active()


# class AudioBufferManager:
#     def __init__(self, audio_buffer: queue.Queue):
#         self.audio_buffer = audio_buffer
#         self.total_samples = 0

#     def add_to_buffer(self, audio_data):
#         self.audio_buffer.put(audio_data)
#         self.total_samples += len(audio_data) // 2

#     def clear_buffer(self):
#         while not self.audio_buffer.empty():
#             try:
#                 self.audio_buffer.get_nowait()
#             except queue.Empty:
#                 continue
#         self.total_samples = 0        

#     def get_from_buffer(self, timeout: float = 0.05):
#         try:
#             chunk = self.audio_buffer.get(timeout=timeout)
#             self.total_samples -= len(chunk) // 2
#             return chunk
#         except queue.Empty:
#             return None

#     def get_buffered_seconds(self, rate: int) -> float:
#         return self.total_samples / rate


# class StreamPlayer:
#     def __init__(self, audio_buffer: queue.Queue, config: AudioConfiguration):
#         self.buffer_manager = AudioBufferManager(audio_buffer)
#         self.audio_stream = AudioStream(config)
#         self.playback_active = False
#         self.immediate_stop = threading.Event()
#         self.pause_event = threading.Event()
#         self.playback_thread = None

#     def _play_chunk(self, chunk):
#         sub_chunk_size = 1024
#         for i in range(0, len(chunk), sub_chunk_size):
#             sub_chunk = chunk[i:i + sub_chunk_size]
#             self.audio_stream.stream.write(sub_chunk)

#             # Pause playback if the event is set
#             while self.pause_event.is_set():
#                 time.sleep(0.01)
            
#             if self.immediate_stop.is_set():
#                 break

#     def _process_buffer(self):
#         while self.playback_active or not self.buffer_manager.audio_buffer.empty():
#             chunk = self.buffer_manager.get_from_buffer()
#             if chunk:
#                 self._play_chunk(chunk)

#             if self.immediate_stop.is_set():
#                 logging.info("Immediate stop requested, aborting playback")
#                 break

#     def get_buffered_seconds(self):
#         total_samples = sum(len(chunk) // 2 for chunk in list(self.buffer_manager.audio_buffer.queue))
#         return total_samples / self.audio_stream.config.rate  

#     def start(self):
#         self.playback_active = True
#         self.audio_stream.open_stream()
#         self.audio_stream.start_stream()
#         self.playback_thread = threading.Thread(target=self._process_buffer)
#         self.playback_thread.start()

#     def stop(self, immediate: bool = False):

#         if not self.playback_thread:
#             logging.warn("No playback thread found, cannot stop playback")
#             return

#         if immediate:
#             self.immediate_stop.set()
#             return

#         self.playback_active = False
        
#         if self.playback_thread and self.playback_thread.is_alive():
#             self.playback_thread.join()

#         time.sleep(0.1)

#         self.audio_stream.close_stream()
#         self.immediate_stop.clear()
#         self.buffer_manager.clear_buffer()
#         self.playback_thread = None

#     def pause(self):
#         self.pause_event.set()

#     def resume(self):
#         self.pause_event.clear()



from .engines import BaseEngine
from .threadsafe_generators import CharIterator, AccumulatingThreadSafeGenerator
from .stream_player import StreamPlayer, AudioConfiguration

import stream2sentence as s2s
from typing import Iterator

import threading
import logging
from typing import Union, Iterator


class TextToAudioStream:

    def __init__(self, 
                 engine: BaseEngine,
                 level=logging.WARNING ,
                 ):
        """Initializes the TextToAudioStream.

        Args:
            engine: The engine used for text to audio synthesis.
            level: Logging level.
        """

        # Initialize the logging configuration with the specified level
        logging.basicConfig(format='RealTimeTTS: %(message)s', level=level)

        # Store the engine instance (responsible for text-to-audio conversion)
        self.engine = engine

        # Extract stream information (format, channels, rate) from the engine
        format, channels, rate = self.engine.get_stream_info()

        # Create a CharIterator instance for managing individual characters
        self.char_iter = CharIterator()

        # Create a thread-safe version of the char iterator
        self.thread_safe_char_iter = AccumulatingThreadSafeGenerator(self.char_iter)

        # Check if the engine doesn't support consuming generators directly
        if not self.engine.can_consume_generators:
            
            # Initialize a StreamPlayer for managing audio playback if the engine can't handle generators directly
            self.player = StreamPlayer(self.engine.queue, AudioConfiguration(format, channels, rate))
        
        # Initialize the play_thread attribute (used for playing audio in a separate thread)
        self.play_thread = None

        # Initialize an attribute to store generated text
        self.generated_text = ""

        # A flag to indicate if the audio stream is currently running or not
        self.stream_running = False

    def feed(self, 
             text_or_iterator: Union[str, Iterator[str]]):
        """Feeds text or an iterator to the stream.

        Args:
            text_or_iterator: Text or iterator to be fed.

        Returns:
            Self instance.
        """
        self.char_iter.add(text_or_iterator)
        return self        

    def play_async(self,   
                   fast_sentence_fragment: bool = False,
                   buffer_threshold_seconds: float = 2.0,
                   minimum_sentence_length: int = 3, 
                   log_characters: bool = False,
                   log_synthesized_text = False             
                   ):
        """
        Handles the synthesis of text to audio.
        Initiates the audio stream playback in a separate thread, allowing the main thread to continue without waiting for the playback to finish.
        If the engine can't consume generators, it utilizes a player.

        Args:
        - fast_sentence_fragment: Determines if sentence fragments should be quickly yielded. Useful when a faster response is desired even if a sentence isn't complete.
        - buffer_threshold_seconds: Time in seconds to determine the buffering threshold. Helps to decide when to generate more audio based on buffered content.
        - minimum_sentence_length: Minimum characters required to treat content as a sentence.
        - log_characters: If True, logs the characters processed for synthesis.
        - log_synthesized_text: If True, logs the synthesized text chunks.
        """
        
        self.stream_running = True
        self.play_thread = threading.Thread(target=self.play, args=(fast_sentence_fragment, buffer_threshold_seconds, minimum_sentence_length, log_characters, log_synthesized_text))
        self.play_thread.start()        

    def play(self,
            fast_sentence_fragment: bool = False,
            buffer_threshold_seconds: float = 2.0,
            minimum_sentence_length: int = 3,
            log_characters: bool = False,
            log_synthesized_text = False):
        """
        Handles the synthesis of text to audio.
        Plays the audio stream and waits until it is finished playing.
        If the engine can't consume generators, it utilizes a player.

        Args:
        - fast_sentence_fragment: Determines if sentence fragments should be quickly yielded. Useful when a faster response is desired even if a sentence isn't complete.
        - buffer_threshold_seconds: Time in seconds to determine the buffering threshold. Helps to decide when to generate more audio based on buffered content.
        - minimum_sentence_length: Minimum characters required to treat content as a sentence.
        - log_characters: If True, logs the characters processed for synthesis.
        - log_synthesized_text: If True, logs the synthesized text chunks.
        """

        # Log the start of the stream
        logging.info(f"stream start")

        # Set the stream_running flag to indicate the stream is active
        self.stream_running = True

        # Initialize the generated_text variable
        self.generated_text = ""

        # Check if the engine can handle generators directly
        if self.engine.can_consume_generators:

            try:
                # Directly synthesize audio using the character iterator
                self.engine.synthesize(self.char_iter)

            finally:
                # Once done, set the stream running flag to False and log the stream stop
                self.stream_running = False
                logging.info("stream stop")

                # Accumulate the generated text and reset the character iterators
                self.generated_text = self.thread_safe_char_iter.accumulated_text()

                self.char_iter = CharIterator()
                self.thread_safe_char_iter = AccumulatingThreadSafeGenerator(self.char_iter)
        else:
            try:
                # Start the audio player to handle playback
                self.player.start()

                # Generate sentences from the characters
                generate_sentences = s2s.generate_sentences(self.thread_safe_char_iter, minimum_sentence_length=minimum_sentence_length, quick_yield_single_sentence_fragment=fast_sentence_fragment, cleanup_text_links=True, cleanup_text_emojis=True, log_characters=log_characters)
                
                # Create the synthesis chunk generator with the given sentences
                chunk_generator = self.synthesis_chunk_generator(generate_sentences, buffer_threshold_seconds, log_synthesized_text)

                # Iterate through the synthesized chunks and feed them to the engine for audio synthesis
                for sentence in chunk_generator:

                    sentence = sentence.strip()
                    if log_synthesized_text:
                        logging.info(f"synthesizing: {sentence}")

                    self.engine.synthesize(sentence)

                    # Break out of the loop if the stream is no longer running
                    if not self.stream_running:
                        break
            finally:
                # Once done, stop the player, set the stream running flag to False, and log the stream stop
                self.player.stop()
                self.stream_running = False
                logging.info("stream stop")

                # Accumulate the generated text and reset the character iterators
                self.generated_text = self.thread_safe_char_iter.accumulated_text()

                self.char_iter = CharIterator()
                self.thread_safe_char_iter = AccumulatingThreadSafeGenerator(self.char_iter)

    def pause(self):
        """Pauses playback of the synthesized audio stream (won't work properly with elevenlabs)."""
        if self.is_playing():
            logging.info("stream pause")
            if self.engine.can_consume_generators:
                self.engine.pause()
            else:
                self.player.pause()

    def resume(self):
        """Resumes a previously paused playback of the synthesized audio stream (won't work properly with elevenlabs)."""
        if self.is_playing():
            logging.info("stream resume")
            if self.engine.can_consume_generators:
                self.engine.resume()
            else:
                self.player.resume()

    def stop(self):
        """Stops the playback of the synthesized audio stream immediately."""
        if self.is_playing():
            self.char_iter.stop()
            if self.engine.can_consume_generators:
                if self.engine.stop():
                    self.stream_running = False
            else:
                self.player.stop(immediate=True)
                self.stream_running = False
    
    def text(self):
        """Retrieves the text that has been fed into the stream.
        Returns:
            The accumulated text.
        """        
        if self.generated_text:
            return self.generated_text
        return self.thread_safe_char_iter.accumulated_text()

    def is_playing(self):
        """Checks if the stream is currently playing.
        Returns:
            Boolean indicating if the stream is playing.
        """ 
        return self.stream_running

    def synthesis_chunk_generator(self,
                                  generator: Iterator[str],
                                  buffer_threshold_seconds: float = 2.0,
                                  log_synthesis_chunks: bool = False) -> Iterator[str]:
        """
        Generates synthesis chunks based on buffered audio length.

        The function buffers chunks of synthesis until the buffered audio seconds fall below the provided threshold. 
        Once the threshold is crossed, the buffered synthesis chunk is yielded.

        Args:
            generator: Input iterator that provides chunks for synthesis.
            buffer_threshold_seconds: Time in seconds to specify how long audio data should be buffered 
                                    before yielding the synthesis chunk.
            log_synthesis_chunks: Boolean flag that, if set to True, logs the synthesis chunks to the logging system.

        Returns:
            Iterator of synthesis chunks.
        """     

        # Initializes an empty string to accumulate chunks of synthesis
        synthesis_chunk = ""
        
        # Iterates over each chunk from the provided generator
        for chunk in generator:

            # Fetch the total seconds of buffered audio
            buffered_audio_seconds = self.player.get_buffered_seconds()
            
            # Append the current chunk (and a space) to the accumulated synthesis_chunk
            synthesis_chunk += chunk + " "
            
            # Check if the buffered audio is below the specified threshold
            if buffered_audio_seconds < buffer_threshold_seconds:
                # If the log_synthesis_chunks flag is True, log the current synthesis_chunk
                if log_synthesis_chunks:
                    logging.info(f"-- \"{synthesis_chunk}\"")
                
                # Yield the current synthesis_chunk and reset it for the next set of accumulations
                yield synthesis_chunk
                synthesis_chunk = ""

        # After iterating over all chunks, check if there's any remaining data in synthesis_chunk
        if synthesis_chunk:
            # If the log_synthesis_chunks flag is True, log the remaining synthesis_chunk
            if log_synthesis_chunks:
                logging.info(f"-- \"{synthesis_chunk}\"")
            
            # Yield the remaining synthesis_chunk
            yield synthesis_chunk